from __future__ import print_function

import __builtin__
import functools
import inspect
import pdb
import signal
import sys
import threading
import time
import traceback as tb

import Globals

def stacktrace(func):
	@functools.wraps(func)
	def wrapped(*args, **kwargs):
		print_bt(func.__name__)
		return func(*args, **kwargs)

	return wrapped


def print_bt(name=None):
	if name is not None:
		dprint('=== {} backtrace:'.format(name))

	for line in tb.format_stack()[:-1]:
		dprint(line.strip())

	if name is not None:
		dprint('=== end {} backtrace'.format(name))


class DebugPrinter:
	lock = threading.RLock()

	@classmethod
	def _print(cls, *args, **kwargs):
		prefix = kwargs.pop('prefix', False)

		if prefix:
			prefix = 'SERVER' if Globals.server else 'CLIENT'
			print('[{}] ', end='', file=sys.stderr)

		print(*args, file=sys.stderr)
		sys.stderr.flush()


def dprint(*args, **kwargs):
	with DebugPrinter.lock:
		DebugPrinter._print(*args, **kwargs)

def dlock():
	return DebugPrinter.lock

def replace_builtin(name):
	def wrapper(f):
		original = getattr(__builtin__, name)

		@functools.wraps(f)
		def wrapped(*args, **kwargs):
			return f(original, *args, **kwargs)

		setattr(__builtin__, name, wrapped)
		return wrapped
	return wrapper


class count_iter:
	def __init__(self, iterable):
		self.iterable = iterable
		self.length = 0

	def __iter__(self):
		return self

	def next(self):
		item = self.iterable.next()
		self.length += 1
		return item


def print_iter(iterable):
	for item in iterable:
		dprint(str(item))
		yield item


def debug_iter(iterable):
	for item in iterable:
		yield item
		time.sleep(1)


#  @replace_builtin('open')
def debug_open(_open, path, *args, **kwargs):
	f = _open(path, *args, **kwargs)

	return f


class DebugLock(threading._RLock):

	def acquire(self):
		t = threading.current_thread()
		#  dprint('{} lock acquire'.format(t.name), prefix=True)
		print_bt('{} lock acquire'.format(t.name))
		return super(DebugLock, self).acquire()

	def release(self):
		t = threading.current_thread()
		dprint('{} released lock'.format(t.name))
		return super(DebugLock, self).release()

#  threading._RLock = DebugLock


class Crasher:
	def __init__(self, count):
		self.count = count
		self.cur = 0

	def ping(self):
		if self.cur <= self.count:
			self.cur += 1
		if self.cur == self.count:
			raise RuntimeError('Crashed!')


def setup_pdb():
	def debug_signal_handler(signal, frame):
		pdb.set_trace()

	signal.signal(signal.SIGUSR1, debug_signal_handler)
